@@ -20,23 +20,32 @@
 package snapstate_test
 
 import (
+	"context"
 	"errors"
 	"fmt"
 	"io"
+	"os"
 	"path/filepath"
 	"sort"
 	"strings"
+	"sync"
 
-	"golang.org/x/net/context"
+	. "gopkg.in/check.v1"
 
+	"github.com/snapcore/snapd/boot"
+	"github.com/snapcore/snapd/cmd/snaplock/runinhibit"
+	"github.com/snapcore/snapd/osutil"
 	"github.com/snapcore/snapd/overlord/auth"
 	"github.com/snapcore/snapd/overlord/snapstate"
 	"github.com/snapcore/snapd/overlord/snapstate/backend"
 	"github.com/snapcore/snapd/overlord/state"
 	"github.com/snapcore/snapd/progress"
 	"github.com/snapcore/snapd/snap"
+	"github.com/snapcore/snapd/snap/snapfile"
 	"github.com/snapcore/snapd/store"
 	"github.com/snapcore/snapd/store/storetest"
+	"github.com/snapcore/snapd/strutil"
+	"github.com/snapcore/snapd/timings"
 )
 
 type fakeOp struct {
@@ -44,6 +53,7 @@ type fakeOp struct {
 
 	name    string
 	channel string
+	path    string
 	revno   snap.Revision
 	sinfo   snap.SideInfo
 	stype   snap.Type
@@ -57,10 +67,33 @@ type fakeOp struct {
 	rmAliases []*backend.Alias
 
 	userID int
+
+	otherInstances         bool
+	unlinkFirstInstallUndo bool
+
+	services         []string
+	disabledServices []string
+
+	vitalityRank int
+
+	inhibitHint runinhibit.Hint
+
+	requireSnapdTooling bool
 }
 
 type fakeOps []fakeOp
 
+func (ops fakeOps) MustFindOp(c *C, opName string) *fakeOp {
+	for _, op := range ops {
+		if op.op == opName {
+			return &op
+		}
+	}
+	c.Errorf("cannot find operation with op: %q, all ops: %v", opName, ops.Ops())
+	c.FailNow()
+	return nil
+}
+
 func (ops fakeOps) Ops() []string {
 	opsOps := make([]string, len(ops))
 	for i, op := range ops {
@@ -93,6 +126,8 @@ func (ops fakeOps) First(op string) *fakeOp {
 type fakeDownload struct {
 	name     string
 	macaroon string
+	target   string
+	opts     *store.DownloadOptions
 }
 
 type byName []store.CurrentSnap
@@ -100,7 +135,7 @@ type byName []store.CurrentSnap
 func (bna byName) Len() int      { return len(bna) }
 func (bna byName) Swap(i, j int) { bna[i], bna[j] = bna[j], bna[i] }
 func (bna byName) Less(i, j int) bool {
-	return bna[i].Name < bna[j].Name
+	return bna[i].InstanceName < bna[j].InstanceName
 }
 
 type byAction []*store.SnapAction
@@ -112,7 +147,7 @@ func (ba byAction) Less(i, j int) bool {
 		if ba[i].Action == "refresh" {
 			return ba[i].SnapID < ba[j].SnapID
 		} else {
-			return ba[i].Name < ba[j].Name
+			return ba[i].InstanceName < ba[j].InstanceName
 		}
 	}
 	return ba[i].Action < ba[j].Action
@@ -127,6 +162,7 @@ type fakeStore struct {
 	fakeCurrentProgress int
 	fakeTotalProgress   int
 	state               *state.State
+	seenPrivacyKeys     map[string]bool
 }
 
 func (f *fakeStore) pokeStateLock() {
@@ -136,33 +172,67 @@ func (f *fakeStore) pokeStateLock() {
 	f.state.Unlock()
 }
 
-func (f *fakeStore) SnapInfo(spec store.SnapSpec, user *auth.UserState) (*snap.Info, error) {
+func (f *fakeStore) SnapInfo(ctx context.Context, spec store.SnapSpec, user *auth.UserState) (*snap.Info, error) {
 	f.pokeStateLock()
 
-	info, err := f.snapInfo(spec, user)
+	_, instanceKey := snap.SplitInstanceName(spec.Name)
+	if instanceKey != "" {
+		return nil, fmt.Errorf("internal error: unexpected instance name: %q", spec.Name)
+	}
+	sspec := snapSpec{
+		Name: spec.Name,
+	}
+	info, err := f.snap(sspec, user)
 
 	userID := 0
 	if user != nil {
 		userID = user.ID
 	}
-	f.fakeBackend.ops = append(f.fakeBackend.ops, fakeOp{op: "storesvc-snap", name: spec.Name, revno: info.Revision, userID: userID})
+	f.fakeBackend.appendOp(&fakeOp{op: "storesvc-snap", name: spec.Name, revno: info.Revision, userID: userID})
 
 	return info, err
 }
 
-func (f *fakeStore) snapInfo(spec store.SnapSpec, user *auth.UserState) (*snap.Info, error) {
+type snapSpec struct {
+	Name     string
+	Channel  string
+	Revision snap.Revision
+	Cohort   string
+}
+
+func (f *fakeStore) snap(spec snapSpec, user *auth.UserState) (*snap.Info, error) {
 	if spec.Revision.Unset() {
-		spec.Revision = snap.R(11)
-		if spec.Channel == "channel-for-7" {
-			spec.Revision.N = 7
+		switch {
+		case spec.Cohort != "":
+			spec.Revision = snap.R(666)
+		case spec.Channel == "channel-for-7":
+			spec.Revision = snap.R(7)
+		default:
+			spec.Revision = snap.R(11)
 		}
 	}
 
 	confinement := snap.StrictConfinement
 
 	typ := snap.TypeApp
-	if spec.Name == "some-core" {
+	epoch := snap.E("1*")
+	switch spec.Name {
+	case "core", "core16", "ubuntu-core", "some-core":
 		typ = snap.TypeOS
+	case "some-base", "other-base", "some-other-base", "yet-another-base", "core18":
+		typ = snap.TypeBase
+	case "some-kernel":
+		typ = snap.TypeKernel
+	case "some-gadget", "brand-gadget":
+		typ = snap.TypeGadget
+	case "some-snapd":
+		typ = snap.TypeSnapd
+	case "snapd":
+		typ = snap.TypeSnapd
+	case "some-snap-now-classic":
+		confinement = "classic"
+	case "some-epoch-snap":
+		epoch = snap.E("42")
 	}
 
 	if spec.Name == "snap-unknown" {
@@ -180,11 +250,15 @@ func (f *fakeStore) snapInfo(spec store.SnapSpec, user *auth.UserState) (*snap.I
 		Version: spec.Name,
 		DownloadInfo: snap.DownloadInfo{
 			DownloadURL: "https://some-server.com/some/path.snap",
+			Size:        5,
 		},
 		Confinement: confinement,
-		Type:        typ,
+		SnapType:    typ,
+		Epoch:       epoch,
 	}
 	switch spec.Channel {
+	case "channel-no-revision":
+		return nil, &store.RevisionNotAvailableError{}
 	case "channel-for-devmode":
 		info.Confinement = snap.DevModeConfinement
 	case "channel-for-classic":
@@ -202,6 +276,39 @@ func (f *fakeStore) snapInfo(spec store.SnapSpec, user *auth.UserState) (*snap.I
 				Symlink: "$SNAP/usr",
 			},
 		}
+	case "channel-for-user-daemon":
+		info.Apps = map[string]*snap.AppInfo{
+			"user-daemon": {
+				Snap:        info,
+				Name:        "user-daemon",
+				Daemon:      "simple",
+				DaemonScope: "user",
+			},
+		}
+	case "channel-for-dbus-activation":
+		slot := &snap.SlotInfo{
+			Snap:      info,
+			Name:      "dbus-slot",
+			Interface: "dbus",
+			Attrs: map[string]interface{}{
+				"bus":  "system",
+				"name": "org.example.Foo",
+			},
+			Apps: make(map[string]*snap.AppInfo),
+		}
+		info.Apps = map[string]*snap.AppInfo{
+			"dbus-daemon": {
+				Snap:        info,
+				Name:        "dbus-daemon",
+				Daemon:      "simple",
+				DaemonScope: snap.SystemDaemon,
+				ActivatesOn: []*snap.SlotInfo{slot},
+				Slots: map[string]*snap.SlotInfo{
+					slot.Name: slot,
+				},
+			},
+		}
+		slot.Apps["dbus-daemon"] = info.Apps["dbus-daemon"]
 	}
 
 	return info, nil
@@ -219,6 +326,7 @@ func (f *fakeStore) lookupRefresh(cand refreshCand) (*snap.Info, error) {
 	var name string
 
 	typ := snap.TypeApp
+	epoch := snap.E("1*")
 	switch cand.snapID {
 	case "":
 		panic("store refresh APIs expect snap-ids")
@@ -230,9 +338,21 @@ func (f *fakeStore) lookupRefresh(cand refreshCand) (*snap.Info, error) {
 		name = "services-snap"
 	case "some-snap-id":
 		name = "some-snap"
+	case "some-other-snap-id":
+		name = "some-other-snap"
+	case "some-epoch-snap-id":
+		name = "some-epoch-snap"
+		epoch = snap.E("42")
+	case "some-snap-now-classic-id":
+		name = "some-snap-now-classic"
+	case "some-snap-was-classic-id":
+		name = "some-snap-was-classic"
 	case "core-snap-id":
 		name = "core"
 		typ = snap.TypeOS
+	case "core18-snap-id":
+		name = "core18"
+		typ = snap.TypeBase
 	case "snap-with-snapd-control-id":
 		name = "snap-with-snapd-control"
 	case "producer-id":
@@ -244,6 +364,19 @@ func (f *fakeStore) lookupRefresh(cand refreshCand) (*snap.Info, error) {
 		typ = snap.TypeBase
 	case "snap-content-plug-id":
 		name = "snap-content-plug"
+	case "snap-content-slot-id":
+		name = "snap-content-slot"
+	case "snapd-snap-id":
+		name = "snapd"
+		typ = snap.TypeSnapd
+	case "kernel-id":
+		name = "kernel"
+		typ = snap.TypeKernel
+	case "brand-gadget-id":
+		name = "brand-gadget"
+		typ = snap.TypeGadget
+	case "alias-snap-id":
+		name = "snap-id"
 	default:
 		panic(fmt.Sprintf("refresh: unknown snap-id: %s", cand.snapID))
 	}
@@ -254,16 +387,19 @@ func (f *fakeStore) lookupRefresh(cand refreshCand) (*snap.Info, error) {
 	}
 	confinement := snap.StrictConfinement
 	switch cand.channel {
-	case "channel-for-7":
+	case "channel-for-7/stable":
 		revno = snap.R(7)
-	case "channel-for-classic":
+	case "channel-for-classic/stable":
 		confinement = snap.ClassicConfinement
-	case "channel-for-devmode":
+	case "channel-for-devmode/stable":
 		confinement = snap.DevModeConfinement
 	}
+	if name == "some-snap-now-classic" {
+		confinement = "classic"
+	}
 
 	info := &snap.Info{
-		Type: typ,
+		SnapType: typ,
 		SideInfo: snap.SideInfo{
 			RealName: name,
 			Channel:  cand.channel,
@@ -276,9 +412,10 @@ func (f *fakeStore) lookupRefresh(cand refreshCand) (*snap.Info, error) {
 		},
 		Confinement:   confinement,
 		Architectures: []string{"all"},
+		Epoch:         epoch,
 	}
 	switch cand.channel {
-	case "channel-for-layout":
+	case "channel-for-layout/stable":
 		info.Layout = map[string]*snap.Layout{
 			"/usr": {
 				Snap:    info,
@@ -286,7 +423,7 @@ func (f *fakeStore) lookupRefresh(cand refreshCand) (*snap.Info, error) {
 				Symlink: "$SNAP/usr",
 			},
 		}
-	case "channel-for-base":
+	case "channel-for-base/stable":
 		info.Base = "some-base"
 	}
 
@@ -308,26 +445,29 @@ func (f *fakeStore) lookupRefresh(cand refreshCand) (*snap.Info, error) {
 	return nil, store.ErrNoUpdateAvailable
 }
 
-func (f *fakeStore) SnapAction(ctx context.Context, currentSnaps []*store.CurrentSnap, actions []*store.SnapAction, user *auth.UserState, opts *store.RefreshOptions) ([]*snap.Info, error) {
+func (f *fakeStore) SnapAction(ctx context.Context, currentSnaps []*store.CurrentSnap, actions []*store.SnapAction, assertQuery store.AssertionQuery, user *auth.UserState, opts *store.RefreshOptions) ([]store.SnapActionResult, []store.AssertionResult, error) {
 	if ctx == nil {
 		panic("context required")
 	}
 	f.pokeStateLock()
+	if assertQuery != nil {
+		panic("no assertion query support")
+	}
 
 	if len(currentSnaps) == 0 && len(actions) == 0 {
-		return nil, nil
+		return nil, nil, nil
 	}
-	if len(actions) > 3 {
+	if len(actions) > 4 {
 		panic("fake SnapAction unexpectedly called with more than 3 actions")
 	}
 
-	curByID := make(map[string]*store.CurrentSnap, len(currentSnaps))
+	curByInstanceName := make(map[string]*store.CurrentSnap, len(currentSnaps))
 	curSnaps := make(byName, len(currentSnaps))
 	for i, cur := range currentSnaps {
-		if cur.Name == "" || cur.SnapID == "" || cur.Revision.Unset() {
-			return nil, fmt.Errorf("internal error: incomplete current snap info")
+		if cur.InstanceName == "" || cur.SnapID == "" || cur.Revision.Unset() {
+			return nil, nil, fmt.Errorf("internal error: incomplete current snap info")
 		}
-		curByID[cur.SnapID] = cur
+		curByInstanceName[cur.InstanceName] = cur
 		curSnaps[i] = *cur
 	}
 	sort.Sort(curSnaps)
@@ -339,7 +479,19 @@ func (f *fakeStore) SnapAction(ctx context.Context, currentSnaps []*store.Curren
 	if len(curSnaps) == 0 {
 		curSnaps = nil
 	}
-	f.fakeBackend.ops = append(f.fakeBackend.ops, fakeOp{op: "storesvc-snap-action", curSnaps: curSnaps, userID: userID})
+	f.fakeBackend.appendOp(&fakeOp{
+		op:       "storesvc-snap-action",
+		curSnaps: curSnaps,
+		userID:   userID,
+	})
+
+	if f.seenPrivacyKeys == nil {
+		// so that checks don't topple over this being uninitialized
+		f.seenPrivacyKeys = make(map[string]bool)
+	}
+	if opts != nil && opts.PrivacyKey != "" {
+		f.seenPrivacyKeys[opts.PrivacyKey] = true
+	}
 
 	sorted := make(byAction, len(actions))
 	copy(sorted, actions)
@@ -347,24 +499,30 @@ func (f *fakeStore) SnapAction(ctx context.Context, currentSnaps []*store.Curren
 
 	refreshErrors := make(map[string]error)
 	installErrors := make(map[string]error)
-	var res []*snap.Info
+	var res []store.SnapActionResult
 	for _, a := range sorted {
 		if a.Action != "install" && a.Action != "refresh" {
 			panic("not supported")
 		}
+		if a.InstanceName == "" {
+			return nil, nil, fmt.Errorf("internal error: action without instance name")
+		}
+
+		snapName, instanceKey := snap.SplitInstanceName(a.InstanceName)
 
 		if a.Action == "install" {
-			spec := store.SnapSpec{
-				Name:     a.Name,
+			spec := snapSpec{
+				Name:     snapName,
 				Channel:  a.Channel,
 				Revision: a.Revision,
+				Cohort:   a.CohortKey,
 			}
-			info, err := f.snapInfo(spec, user)
+			info, err := f.snap(spec, user)
 			if err != nil {
-				installErrors[a.Name] = err
+				installErrors[a.InstanceName] = err
 				continue
 			}
-			f.fakeBackend.ops = append(f.fakeBackend.ops, fakeOp{
+			f.fakeBackend.appendOp(&fakeOp{
 				op:     "storesvc-snap-action:action",
 				action: *a,
 				revno:  info.Revision,
@@ -373,13 +531,21 @@ func (f *fakeStore) SnapAction(ctx context.Context, currentSnaps []*store.Curren
 			if !a.Revision.Unset() {
 				info.Channel = ""
 			}
-			res = append(res, info)
+			info.InstanceKey = instanceKey
+			sar := store.SnapActionResult{Info: info}
+			if strings.HasSuffix(snapName, "-with-default-track") && strutil.ListContains([]string{"stable", "candidate", "beta", "edge"}, a.Channel) {
+				sar.RedirectChannel = "2.0/" + a.Channel
+			}
+			res = append(res, sar)
 			continue
 		}
 
 		// refresh
 
-		cur := curByID[a.SnapID]
+		cur := curByInstanceName[a.InstanceName]
+		if cur == nil {
+			return nil, nil, fmt.Errorf("internal error: no matching current snap for %q", a.InstanceName)
+		}
 		channel := a.Channel
 		if channel == "" {
 			channel = cur.TrackingChannel
@@ -412,16 +578,17 @@ func (f *fakeStore) SnapAction(ctx context.Context, currentSnaps []*store.Curren
 			userID: userID,
 		})
 		if err == store.ErrNoUpdateAvailable {
-			refreshErrors[cur.Name] = err
+			refreshErrors[cur.InstanceName] = err
 			continue
 		}
 		if err != nil {
-			return nil, err
+			return nil, nil, err
 		}
 		if !a.Revision.Unset() {
 			info.Channel = ""
 		}
-		res = append(res, info)
+		info.InstanceKey = instanceKey
+		res = append(res, store.SnapActionResult{Info: info})
 	}
 
 	if len(refreshErrors)+len(installErrors) > 0 || len(res) == 0 {
@@ -431,14 +598,14 @@ func (f *fakeStore) SnapAction(ctx context.Context, currentSnaps []*store.Curren
 		if len(installErrors) == 0 {
 			installErrors = nil
 		}
-		return res, &store.SnapActionError{
+		return res, nil, &store.SnapActionError{
 			NoResults: len(refreshErrors)+len(installErrors)+len(res) == 0,
 			Refresh:   refreshErrors,
 			Install:   installErrors,
 		}
 	}
 
-	return res, nil
+	return res, nil, nil
 }
 
 func (f *fakeStore) SuggestedCurrency() string {
@@ -447,18 +614,27 @@ func (f *fakeStore) SuggestedCurrency() string {
 	return "XTS"
 }
 
-func (f *fakeStore) Download(ctx context.Context, name, targetFn string, snapInfo *snap.DownloadInfo, pb progress.Meter, user *auth.UserState) error {
+func (f *fakeStore) Download(ctx context.Context, name, targetFn string, snapInfo *snap.DownloadInfo, pb progress.Meter, user *auth.UserState, dlOpts *store.DownloadOptions) error {
 	f.pokeStateLock()
 
+	if _, key := snap.SplitInstanceName(name); key != "" {
+		return fmt.Errorf("internal error: unsupported download with instance name %q", name)
+	}
 	var macaroon string
 	if user != nil {
 		macaroon = user.StoreMacaroon
 	}
+	// only add the options if they contain anything interesting
+	if *dlOpts == (store.DownloadOptions{}) {
+		dlOpts = nil
+	}
 	f.downloads = append(f.downloads, fakeDownload{
 		macaroon: macaroon,
 		name:     name,
+		target:   targetFn,
+		opts:     dlOpts,
 	})
-	f.fakeBackend.ops = append(f.fakeBackend.ops, fakeOp{op: "storesvc-download", name: name})
+	f.fakeBackend.appendOp(&fakeOp{op: "storesvc-download", name: name})
 
 	pb.SetTotal(float64(f.fakeTotalProgress))
 	pb.Set(float64(f.fakeCurrentProgress))
@@ -471,7 +647,8 @@ func (f *fakeStore) WriteCatalogs(ctx context.Context, _ io.Writer, _ store.Snap
 		panic("context required")
 	}
 	f.pokeStateLock()
-	f.fakeBackend.ops = append(f.fakeBackend.ops, fakeOp{
+
+	f.fakeBackend.appendOp(&fakeOp{
 		op: "x-commands",
 	})
 
@@ -483,7 +660,8 @@ func (f *fakeStore) Sections(ctx context.Context, _ *auth.UserState) ([]string,
 		panic("context required")
 	}
 	f.pokeStateLock()
-	f.fakeBackend.ops = append(f.fakeBackend.ops, fakeOp{
+
+	f.fakeBackend.appendOp(&fakeOp{
 		op: "x-sections",
 	})
 
@@ -492,40 +670,96 @@ func (f *fakeStore) Sections(ctx context.Context, _ *auth.UserState) ([]string,
 
 type fakeSnappyBackend struct {
 	ops fakeOps
+	mu  sync.Mutex
+
+	linkSnapWaitCh      chan int
+	linkSnapWaitTrigger string
+	linkSnapFailTrigger string
+	linkSnapMaybeReboot bool
 
-	linkSnapFailTrigger     string
 	copySnapDataFailTrigger string
 	emptyContainer          snap.Container
+
+	servicesCurrentlyDisabled []string
+
+	lockDir string
 }
 
 func (f *fakeSnappyBackend) OpenSnapFile(snapFilePath string, si *snap.SideInfo) (*snap.Info, snap.Container, error) {
 	op := fakeOp{
 		op:   "open-snap-file",
-		name: snapFilePath,
+		path: snapFilePath,
 	}
 
 	if si != nil {
 		op.sinfo = *si
 	}
 
-	name := filepath.Base(snapFilePath)
-	if idx := strings.IndexByte(name, '_'); idx > -1 {
-		name = name[:idx]
+	var info *snap.Info
+	if !osutil.IsDirectory(snapFilePath) {
+		name := filepath.Base(snapFilePath)
+		split := strings.Split(name, "_")
+		if len(split) >= 2 {
+			// <snap>_<rev>.snap
+			// <snap>_<instance-key>_<rev>.snap
+			name = split[0]
+		}
+
+		info = &snap.Info{SuggestedName: name, Architectures: []string{"all"}}
+		if name == "some-snap-now-classic" {
+			info.Confinement = "classic"
+		}
+		if name == "some-epoch-snap" {
+			info.Epoch = snap.E("42")
+		} else {
+			info.Epoch = snap.E("1*")
+		}
+	} else {
+		// for snap try only
+		snapf, err := snapfile.Open(snapFilePath)
+		if err != nil {
+			return nil, nil, err
+		}
+
+		info, err = snap.ReadInfoFromSnapFile(snapf, si)
+		if err != nil {
+			return nil, nil, err
+		}
 	}
 
-	f.ops = append(f.ops, op)
-	return &snap.Info{SuggestedName: name, Architectures: []string{"all"}}, f.emptyContainer, nil
+	if info == nil {
+		return nil, nil, fmt.Errorf("internal error: no mocked snap for %q", snapFilePath)
+	}
+	f.appendOp(&op)
+	return info, f.emptyContainer, nil
 }
 
-func (f *fakeSnappyBackend) SetupSnap(snapFilePath string, si *snap.SideInfo, p progress.Meter) (snap.Type, error) {
+// XXX: this is now something that is overridden by tests that need a
+//      different service setup so it should be configurable and part
+//      of the fakeSnappyBackend?
+var servicesSnapYaml = `name: services-snap
+apps:
+  svc1:
+    daemon: simple
+    before: [svc3]
+  svc2:
+    daemon: simple
+    after: [svc1]
+  svc3:
+    daemon: simple
+    before: [svc2]
+`
+
+func (f *fakeSnappyBackend) SetupSnap(snapFilePath, instanceName string, si *snap.SideInfo, dev boot.Device, p progress.Meter) (snap.Type, *backend.InstallRecord, error) {
 	p.Notify("setup-snap")
 	revno := snap.R(0)
 	if si != nil {
 		revno = si.Revision
 	}
-	f.ops = append(f.ops, fakeOp{
+	f.appendOp(&fakeOp{
 		op:    "setup-snap",
-		name:  snapFilePath,
+		name:  instanceName,
+		path:  snapFilePath,
 		revno: revno,
 	})
 	snapType := snap.TypeApp
@@ -535,40 +769,56 @@ func (f *fakeSnappyBackend) SetupSnap(snapFilePath string, si *snap.SideInfo, p
 	case "gadget":
 		snapType = snap.TypeGadget
 	}
-	return snapType, nil
+	if instanceName == "borken-in-setup" {
+		return snapType, nil, fmt.Errorf("cannot install snap %q", instanceName)
+	}
+	if instanceName == "some-snap-no-install-record" {
+		return snapType, nil, nil
+	}
+	return snapType, &backend.InstallRecord{}, nil
 }
 
 func (f *fakeSnappyBackend) ReadInfo(name string, si *snap.SideInfo) (*snap.Info, error) {
 	if name == "borken" && si.Revision == snap.R(2) {
 		return nil, errors.New(`cannot read info for "borken" snap`)
 	}
+	if name == "borken-undo-setup" && si.Revision == snap.R(2) {
+		return nil, errors.New(`cannot read info for "borken-undo-setup" snap`)
+	}
 	if name == "not-there" && si.Revision == snap.R(2) {
 		return nil, &snap.NotFoundError{Snap: name, Revision: si.Revision}
 	}
+	snapName, instanceKey := snap.SplitInstanceName(name)
 	// naive emulation for now, always works
 	info := &snap.Info{
-		SuggestedName: name,
+		SuggestedName: snapName,
 		SideInfo:      *si,
 		Architectures: []string{"all"},
-		Type:          snap.TypeApp,
+		SnapType:      snap.TypeApp,
+		Epoch:         snap.E("1*"),
 	}
-	if strings.Contains(name, "alias-snap") {
-		name = "alias-snap"
+	if strings.Contains(snapName, "alias-snap") {
+		// only for the switch below
+		snapName = "alias-snap"
 	}
-	switch name {
-	case "gadget":
-		info.Type = snap.TypeGadget
+	switch snapName {
+	case "snap-with-empty-epoch":
+		info.Epoch = snap.Epoch{}
+	case "some-epoch-snap":
+		info.Epoch = snap.E("13")
+	case "some-snap-with-base":
+		info.Base = "core18"
+	case "gadget", "brand-gadget":
+		info.SnapType = snap.TypeGadget
 	case "core":
-		info.Type = snap.TypeOS
+		info.SnapType = snap.TypeOS
+	case "snapd":
+		info.SnapType = snap.TypeSnapd
 	case "services-snap":
 		var err error
-		info, err = snap.InfoFromSnapYaml([]byte(`name: services-snap
-apps:
-  svc1:
-    daemon: simple
-  svc2:
-    daemon: simple
-`))
+		// fix services after/before so that there is only one solution
+		// to dependency ordering
+		info, err = snap.InfoFromSnapYaml([]byte(servicesSnapYaml))
 		if err != nil {
 			panic(err)
 		}
@@ -591,13 +841,14 @@ apps:
 		info.SideInfo = *si
 	}
 
+	info.InstanceKey = instanceKey
 	return info, nil
 }
 
 func (f *fakeSnappyBackend) ClearTrashedData(si *snap.Info) {
-	f.ops = append(f.ops, fakeOp{
+	f.appendOp(&fakeOp{
 		op:    "cleanup-trash",
-		name:  si.Name(),
+		name:  si.InstanceName(),
 		revno: si.Revision,
 	})
 }
@@ -616,36 +867,50 @@ func (f *fakeSnappyBackend) CopySnapData(newInfo, oldInfo *snap.Info, p progress
 	}
 
 	if newInfo.MountDir() == f.copySnapDataFailTrigger {
-		f.ops = append(f.ops, fakeOp{
+		f.appendOp(&fakeOp{
 			op:   "copy-data.failed",
-			name: newInfo.MountDir(),
+			path: newInfo.MountDir(),
 			old:  old,
 		})
 		return errors.New("fail")
 	}
 
-	f.ops = append(f.ops, fakeOp{
+	f.appendOp(&fakeOp{
 		op:   "copy-data",
-		name: newInfo.MountDir(),
+		path: newInfo.MountDir(),
 		old:  old,
 	})
 	return nil
 }
 
-func (f *fakeSnappyBackend) LinkSnap(info *snap.Info) error {
-	if info.MountDir() == f.linkSnapFailTrigger {
-		f.ops = append(f.ops, fakeOp{
-			op:   "link-snap.failed",
-			name: info.MountDir(),
-		})
-		return errors.New("fail")
+func (f *fakeSnappyBackend) LinkSnap(info *snap.Info, dev boot.Device, linkCtx backend.LinkContext, tm timings.Measurer) (rebootRequired bool, err error) {
+	if info.MountDir() == f.linkSnapWaitTrigger {
+		f.linkSnapWaitCh <- 1
+		<-f.linkSnapWaitCh
 	}
 
-	f.ops = append(f.ops, fakeOp{
+	op := fakeOp{
 		op:   "link-snap",
-		name: info.MountDir(),
-	})
-	return nil
+		path: info.MountDir(),
+
+		vitalityRank:        linkCtx.VitalityRank,
+		requireSnapdTooling: linkCtx.RequireMountedSnapdSnap,
+	}
+
+	if info.MountDir() == f.linkSnapFailTrigger {
+		op.op = "link-snap.failed"
+		f.ops = append(f.ops, op)
+		return false, errors.New("fail")
+	}
+
+	f.appendOp(&op)
+
+	reboot := false
+	if f.linkSnapMaybeReboot {
+		reboot = info.InstanceName() == dev.Base()
+	}
+
+	return reboot, nil
 }
 
 func svcSnapMountDir(svcs []*snap.AppInfo) string {
@@ -658,29 +923,76 @@ func svcSnapMountDir(svcs []*snap.AppInfo) string {
 	return svcs[0].Snap.MountDir()
 }
 
-func (f *fakeSnappyBackend) StartServices(svcs []*snap.AppInfo, meter progress.Meter) error {
-	f.ops = append(f.ops, fakeOp{
-		op:   "start-snap-services",
-		name: svcSnapMountDir(svcs),
-	})
+func (f *fakeSnappyBackend) StartServices(svcs []*snap.AppInfo, disabledSvcs []string, meter progress.Meter, tm timings.Measurer) error {
+	services := make([]string, 0, len(svcs))
+	for _, svc := range svcs {
+		services = append(services, svc.Name)
+	}
+	op := fakeOp{
+		op:       "start-snap-services",
+		path:     svcSnapMountDir(svcs),
+		services: services,
+	}
+	// only add the services to the op if there's something to add
+	if len(disabledSvcs) != 0 {
+		op.disabledServices = disabledSvcs
+	}
+	f.appendOp(&op)
 	return nil
 }
 
-func (f *fakeSnappyBackend) StopServices(svcs []*snap.AppInfo, reason snap.ServiceStopReason, meter progress.Meter) error {
-	f.ops = append(f.ops, fakeOp{
+func (f *fakeSnappyBackend) StopServices(svcs []*snap.AppInfo, reason snap.ServiceStopReason, meter progress.Meter, tm timings.Measurer) error {
+	f.appendOp(&fakeOp{
 		op:   fmt.Sprintf("stop-snap-services:%s", reason),
-		name: svcSnapMountDir(svcs),
+		path: svcSnapMountDir(svcs),
 	})
 	return nil
 }
 
-func (f *fakeSnappyBackend) UndoSetupSnap(s snap.PlaceInfo, typ snap.Type, p progress.Meter) error {
+func (f *fakeSnappyBackend) ServicesEnableState(info *snap.Info, meter progress.Meter) (map[string]bool, error) {
+	// return the disabled services as disabled and nothing else
+	m := make(map[string]bool)
+	for _, svc := range f.servicesCurrentlyDisabled {
+		m[svc] = false
+	}
+
+	f.appendOp(&fakeOp{
+		op:               "current-snap-service-states",
+		disabledServices: f.servicesCurrentlyDisabled,
+	})
+
+	return m, nil
+}
+
+func (f *fakeSnappyBackend) QueryDisabledServices(info *snap.Info, meter progress.Meter) ([]string, error) {
+	var l []string
+
+	m, err := f.ServicesEnableState(info, meter)
+	if err != nil {
+		return nil, err
+	}
+	for name, enabled := range m {
+		if !enabled {
+			l = append(l, name)
+		}
+	}
+
+	// XXX: add a fakeOp here?
+
+	return l, nil
+}
+
+func (f *fakeSnappyBackend) UndoSetupSnap(s snap.PlaceInfo, typ snap.Type, installRecord *backend.InstallRecord, dev boot.Device, p progress.Meter) error {
 	p.Notify("setup-snap")
-	f.ops = append(f.ops, fakeOp{
+	f.appendOp(&fakeOp{
 		op:    "undo-setup-snap",
-		name:  s.MountDir(),
+		name:  s.InstanceName(),
+		path:  s.MountDir(),
 		stype: typ,
 	})
+	if s.InstanceName() == "borken-undo-setup" {
+		return errors.New(`cannot undo setup of "borken-undo-setup" snap`)
+	}
 	return nil
 }
 
@@ -690,63 +1002,93 @@ func (f *fakeSnappyBackend) UndoCopySnapData(newInfo *snap.Info, oldInfo *snap.I
 	if oldInfo != nil {
 		old = oldInfo.MountDir()
 	}
-	f.ops = append(f.ops, fakeOp{
+	f.appendOp(&fakeOp{
 		op:   "undo-copy-snap-data",
-		name: newInfo.MountDir(),
+		path: newInfo.MountDir(),
 		old:  old,
 	})
 	return nil
 }
 
-func (f *fakeSnappyBackend) UnlinkSnap(info *snap.Info, meter progress.Meter) error {
+func (f *fakeSnappyBackend) UnlinkSnap(info *snap.Info, linkCtx backend.LinkContext, meter progress.Meter) error {
 	meter.Notify("unlink")
-	f.ops = append(f.ops, fakeOp{
+	f.appendOp(&fakeOp{
 		op:   "unlink-snap",
-		name: info.MountDir(),
+		path: info.MountDir(),
+
+		unlinkFirstInstallUndo: linkCtx.FirstInstall,
 	})
 	return nil
 }
 
-func (f *fakeSnappyBackend) RemoveSnapFiles(s snap.PlaceInfo, typ snap.Type, meter progress.Meter) error {
+func (f *fakeSnappyBackend) RemoveSnapFiles(s snap.PlaceInfo, typ snap.Type, installRecord *backend.InstallRecord, dev boot.Device, meter progress.Meter) error {
 	meter.Notify("remove-snap-files")
-	f.ops = append(f.ops, fakeOp{
+	f.appendOp(&fakeOp{
 		op:    "remove-snap-files",
-		name:  s.MountDir(),
+		path:  s.MountDir(),
 		stype: typ,
 	})
 	return nil
 }
 
 func (f *fakeSnappyBackend) RemoveSnapData(info *snap.Info) error {
-	f.ops = append(f.ops, fakeOp{
+	f.appendOp(&fakeOp{
 		op:   "remove-snap-data",
-		name: info.MountDir(),
+		path: info.MountDir(),
 	})
 	return nil
 }
 
 func (f *fakeSnappyBackend) RemoveSnapCommonData(info *snap.Info) error {
-	f.ops = append(f.ops, fakeOp{
+	f.appendOp(&fakeOp{
 		op:   "remove-snap-common-data",
-		name: info.MountDir(),
+		path: info.MountDir(),
+	})
+	return nil
+}
+
+func (f *fakeSnappyBackend) RemoveSnapDataDir(info *snap.Info, otherInstances bool) error {
+	f.ops = append(f.ops, fakeOp{
+		op:             "remove-snap-data-dir",
+		name:           info.InstanceName(),
+		path:           snap.BaseDataDir(info.SnapName()),
+		otherInstances: otherInstances,
+	})
+	return nil
+}
+
+func (f *fakeSnappyBackend) RemoveSnapDir(s snap.PlaceInfo, otherInstances bool) error {
+	f.ops = append(f.ops, fakeOp{
+		op:             "remove-snap-dir",
+		name:           s.InstanceName(),
+		path:           snap.BaseDir(s.SnapName()),
+		otherInstances: otherInstances,
 	})
 	return nil
 }
 
 func (f *fakeSnappyBackend) DiscardSnapNamespace(snapName string) error {
-	f.ops = append(f.ops, fakeOp{
+	f.appendOp(&fakeOp{
 		op:   "discard-namespace",
 		name: snapName,
 	})
 	return nil
 }
 
+func (f *fakeSnappyBackend) RemoveSnapInhibitLock(snapName string) error {
+	f.appendOp(&fakeOp{
+		op:   "remove-inhibit-lock",
+		name: snapName,
+	})
+	return nil
+}
+
 func (f *fakeSnappyBackend) Candidate(sideInfo *snap.SideInfo) {
 	var sinfo snap.SideInfo
 	if sideInfo != nil {
 		sinfo = *sideInfo
 	}
-	f.ops = append(f.ops, fakeOp{
+	f.appendOp(&fakeOp{
 		op:    "candidate",
 		sinfo: sinfo,
 	})
@@ -757,16 +1099,16 @@ func (f *fakeSnappyBackend) CurrentInfo(curInfo *snap.Info) {
 	if curInfo != nil {
 		old = curInfo.MountDir()
 	}
-	f.ops = append(f.ops, fakeOp{
+	f.appendOp(&fakeOp{
 		op:  "current",
 		old: old,
 	})
 }
 
 func (f *fakeSnappyBackend) ForeignTask(kind string, status state.Status, snapsup *snapstate.SnapSetup) {
-	f.ops = append(f.ops, fakeOp{
+	f.appendOp(&fakeOp{
 		op:    kind + ":" + status.String(),
-		name:  snapsup.Name(),
+		name:  snapsup.InstanceName(),
 		revno: snapsup.Revision(),
 	})
 }
@@ -788,7 +1130,7 @@ func (f *fakeSnappyBackend) UpdateAliases(add []*backend.Alias, remove []*backen
 		remove = append([]*backend.Alias(nil), remove...)
 		sort.Sort(byAlias(remove))
 	}
-	f.ops = append(f.ops, fakeOp{
+	f.appendOp(&fakeOp{
 		op:        "update-aliases",
 		aliases:   add,
 		rmAliases: remove,
@@ -797,9 +1139,31 @@ func (f *fakeSnappyBackend) UpdateAliases(add []*backend.Alias, remove []*backen
 }
 
 func (f *fakeSnappyBackend) RemoveSnapAliases(snapName string) error {
-	f.ops = append(f.ops, fakeOp{
+	f.appendOp(&fakeOp{
 		op:   "remove-snap-aliases",
 		name: snapName,
 	})
 	return nil
 }
+
+func (f *fakeSnappyBackend) RunInhibitSnapForUnlink(info *snap.Info, hint runinhibit.Hint, decision func() error) (lock *osutil.FileLock, err error) {
+	f.appendOp(&fakeOp{
+		op:          "run-inhibit-snap-for-unlink",
+		name:        info.InstanceName(),
+		inhibitHint: hint,
+	})
+	if err := decision(); err != nil {
+		return nil, err
+	}
+	if f.lockDir == "" {
+		f.lockDir = os.TempDir()
+	}
+	// XXX: returning a real lock is somewhat annoying
+	return osutil.NewFileLock(filepath.Join(f.lockDir, info.InstanceName()+".lock"))
+}
+
+func (f *fakeSnappyBackend) appendOp(op *fakeOp) {
+	f.mu.Lock()
+	defer f.mu.Unlock()
+	f.ops = append(f.ops, *op)
+}
