@@ -4,94 +4,51 @@
 . "$TESTSLIB/quiet.sh"
 
 debian_name_package() {
+    #shellcheck source=tests/lib/tools/tests.pkgs.apt.sh
+    . "$TESTSLIB/tools/tests.pkgs.apt.sh"
     for i in "$@"; do
-        case "$i" in
-            xdelta3|curl|python3-yaml|kpartx|busybox-static|nfs-kernel-server)
-                echo "$i"
-                ;;
-            man)
-                echo "man-db"
-                ;;
-            *)
-                echo "$i"
-                ;;
-        esac
+        remap_one "$i"
     done
 }
 
 ubuntu_14_04_name_package() {
+    #shellcheck source=tests/lib/tools/tests.pkgs.apt.sh
+    . "$TESTSLIB/tools/tests.pkgs.apt.sh"
     for i in "$@"; do
-        case "$i" in
-            printer-driver-cups-pdf)
-                echo "cups-pdf"
-                ;;
-            *)
-                debian_name_package "$i"
-                ;;
-        esac
+        remap_one "$i"
     done
 }
 
 fedora_name_package() {
+    #shellcheck source=tests/lib/tools/tests.pkgs.dnf-yum.sh
+    . "$TESTSLIB/tools/tests.pkgs.dnf-yum.sh"
     for i in "$@"; do
-        case "$i" in
-            xdelta3|jq|curl|python3-yaml)
-                echo "$i"
-                ;;
-            openvswitch-switch)
-                echo "openvswitch"
-                ;;
-            printer-driver-cups-pdf)
-                echo "cups-pdf"
-                ;;
-            *)
-                echo "$i"
-                ;;
-        esac
+        remap_one "$i"
+    done
+}
+
+amazon_name_package() {
+    #shellcheck source=tests/lib/tools/tests.pkgs.dnf-yum.sh
+    . "$TESTSLIB/tools/tests.pkgs.dnf-yum.sh"
+    for i in "$@"; do
+        remap_one "$i"
     done
 }
 
 opensuse_name_package() {
+    #shellcheck source=tests/lib/tools/tests.pkgs.zypper.sh
+    . "$TESTSLIB/tools/tests.pkgs.zypper.sh"
     for i in "$@"; do
-        case "$i" in
-            python3-yaml)
-                echo "python3-PyYAML"
-                ;;
-            dbus-x11)
-                echo "dbus-1-x11"
-                ;;
-            printer-driver-cups-pdf)
-                echo "cups-pdf"
-                ;;
-            *)
-                echo "$i"
-                ;;
-        esac
+        remap_one "$i"
     done
 }
 
 arch_name_package() {
-    case "$1" in
-        python3-yaml)
-            echo "python-yaml"
-            ;;
-        dbus-x11)
-            # no separate dbus-x11 package in arch
-            echo "dbus"
-            ;;
-        printer-driver-cups-pdf)
-            echo "cups-pdf"
-            ;;
-        openvswitch-switch)
-            echo "openvswitch"
-            ;;
-        man)
-            echo "man-db"
-            ;;
-        *)
-            echo "$1"
-            ;;
-    esac
+    #shellcheck source=tests/lib/tools/tests.pkgs.pacman.sh
+    . "$TESTSLIB/tools/tests.pkgs.pacman.sh"
+    for i in "$@"; do
+        remap_one "$i"
+    done
 }
 
 distro_name_package() {
@@ -102,7 +59,10 @@ distro_name_package() {
         ubuntu-*|debian-*)
             debian_name_package "$@"
             ;;
-        fedora-*)
+        amazon-*|centos-7-*)
+            amazon_name_package "$@"
+            ;;
+        fedora-*|centos-*)
             fedora_name_package "$@"
             ;;
         opensuse-*)
@@ -134,22 +94,25 @@ distro_install_local_package() {
     case "$SPREAD_SYSTEM" in
         ubuntu-14.04-*|debian-*)
             # relying on dpkg as apt(-get) does not support installation from local files in trusty.
-            dpkg -i --force-depends --auto-deconfigure --force-depends-version "$@"
-            apt-get -f install -y
+            eatmydata dpkg -i --force-depends --auto-deconfigure --force-depends-version "$@"
+            eatmydata apt-get -f install -y
             ;;
         ubuntu-*)
-            flags="-y"
+            flags="-y --no-install-recommends"
             if [ "$allow_downgrades" = "true" ]; then
                 flags="$flags --allow-downgrades"
             fi
             # shellcheck disable=SC2086
             apt install $flags "$@"
             ;;
-        fedora-*)
-            quiet dnf -y install "$@"
+        amazon-*|centos-7-*)
+            quiet yum -y localinstall "$@"
+            ;;
+        fedora-*|centos-*)
+            quiet dnf -y install --setopt=install_weak_deps=False "$@"
             ;;
         opensuse-*)
-            quiet rpm -i "$@"
+            quiet zypper in -y --no-recommends --allow-unsigned-rpm "$@"
             ;;
         arch-*)
             pacman -U --noconfirm "$@"
@@ -162,11 +125,27 @@ distro_install_local_package() {
 }
 
 distro_install_package() {
+    orig_xtrace=$(set -o | awk '/xtrace / { print $2 }')
+    set +x
+    echo "distro_install_package $*"
     # Parse additional arguments; once we find the first unknown
     # part we break argument parsing and process all further
     # arguments as package names.
     APT_FLAGS=
     DNF_FLAGS=
+    if os.query is-fedora; then
+        # Fedora images we use come with a number of preinstalled package, among
+        # them gtk3. Those packages are needed to run the tests. The
+        # xdg-desktop-portal-gtk package uses this in the spec:
+        #
+        #   Supplements:    (gtk3 and (flatpak or snapd))
+        #
+        # As a result, when snapd is installed, we will unintentionally pull in
+        # xdg-desktop-portal-gtk and its dependencies breaking tests. For this
+        # reason, disable weak deps altogether.
+        DNF_FLAGS="--setopt=install_weak_deps=False"
+    fi
+    YUM_FLAGS=
     ZYPPER_FLAGS=
     while [ -n "$1" ]; do
         case "$1" in
@@ -174,6 +153,7 @@ distro_install_package() {
                 APT_FLAGS="$APT_FLAGS --no-install-recommends"
                 DNF_FLAGS="$DNF_FLAGS --setopt=install_weak_deps=False"
                 ZYPPER_FLAGS="$ZYPPER_FLAGS --no-recommends"
+                # TODO no way to set this for yum?
                 shift
                 ;;
             *)
@@ -187,21 +167,22 @@ distro_install_package() {
     case "$SPREAD_SYSTEM" in
         ubuntu-*|debian-*)
         if [[ "$*" =~ "libudev-dev" ]]; then
-            apt-get install -y --only-upgrade systemd
+            eatmydata apt-get install -y --only-upgrade systemd
         fi
         ;;
     esac
 
-    # fix dependency issue where libp11-kit0 needs to be downgraded to 
+    # fix dependency issue where libp11-kit0 needs to be downgraded to
     # install gnome-keyring
     case "$SPREAD_SYSTEM" in
         debian-9-*)
         if [[ "$*" =~ "gnome-keyring" ]]; then
-            apt-get remove -y libp11-kit0
+            eatmydata apt-get remove -y libp11-kit0
         fi
         ;;
     esac
 
+    # shellcheck disable=SC2207
     pkg_names=($(
         for pkg in "$@" ; do
             package_name=$(distro_name_package "$pkg")
@@ -217,15 +198,27 @@ distro_install_package() {
     case "$SPREAD_SYSTEM" in
         ubuntu-*|debian-*)
             # shellcheck disable=SC2086
-            quiet apt-get install $APT_FLAGS -y "${pkg_names[@]}"
+            quiet eatmydata apt-get install $APT_FLAGS -y "${pkg_names[@]}"
             ;;
-        fedora-*)
+        amazon-*|centos-7-*)
+            # shellcheck disable=SC2086
+            quiet yum -y install $YUM_FLAGS "${pkg_names[@]}"
+            ;;
+        fedora-*|centos-*)
             # shellcheck disable=SC2086
             quiet dnf -y --refresh install $DNF_FLAGS "${pkg_names[@]}"
             ;;
         opensuse-*)
+            # packages may be downgraded in the repositories, which would be
+            # picked up next time we ran `zypper dup` and applied locally;
+            # however we only update the images periodically, in the meantime,
+            # when downgrades affect packages we need or have installed, `zypper
+            # in` may stop with the prompt asking the user about either breaking
+            # the installed packages or allowing downgrades, passing
+            # --allow-downgrade will make the installation proceed
+
             # shellcheck disable=SC2086
-            quiet zypper install -y $ZYPPER_FLAGS "${pkg_names[@]}"
+            quiet zypper install -y --allow-downgrade --force-resolution $ZYPPER_FLAGS "${pkg_names[@]}"
             ;;
         arch-*)
             # shellcheck disable=SC2086
@@ -236,6 +229,7 @@ distro_install_package() {
             exit 1
             ;;
     esac
+    test "$orig_xtrace" = on && set -x
 }
 
 distro_purge_package() {
@@ -254,9 +248,14 @@ distro_purge_package() {
 
     case "$SPREAD_SYSTEM" in
         ubuntu-*|debian-*)
-            quiet apt-get remove -y --purge -y "$@"
+            # TODO reenable quiet once we have dealt with files being left
+            # behind while purging in prepare
+            eatmydata apt-get remove -y --purge -y "$@"
             ;;
-        fedora-*)
+        amazon-*|centos-7-*)
+            quiet yum -y remove "$@"
+            ;;
+        fedora-*|centos-*)
             quiet dnf -y remove "$@"
             quiet dnf clean all
             ;;
@@ -276,9 +275,13 @@ distro_purge_package() {
 distro_update_package_db() {
     case "$SPREAD_SYSTEM" in
         ubuntu-*|debian-*)
-            quiet apt-get update
+            quiet eatmydata apt-get update
             ;;
-        fedora-*)
+        amazon-*|centos-7-*)
+            quiet yum clean all
+            quiet yum makecache
+            ;;
+        fedora-*|centos-*)
             quiet dnf clean all
             quiet dnf makecache
             ;;
@@ -298,7 +301,13 @@ distro_update_package_db() {
 distro_clean_package_cache() {
     case "$SPREAD_SYSTEM" in
         ubuntu-*|debian-*)
-            quiet apt-get clean
+            quiet eatmydata apt-get clean
+            ;;
+        amazon-*|centos-7-*)
+            yum clean all
+            ;;
+        fedora-*|centos-*)
+            dnf clean all
             ;;
         opensuse-*)
             zypper -q clean --all
@@ -316,9 +325,12 @@ distro_clean_package_cache() {
 distro_auto_remove_packages() {
     case "$SPREAD_SYSTEM" in
         ubuntu-*|debian-*)
-            quiet apt-get -y autoremove
+            quiet eatmydata apt-get -y autoremove
             ;;
-        fedora-*)
+        amazon-*|centos-7-*)
+            quiet yum -y autoremove
+            ;;
+        fedora-*|centos-*)
             quiet dnf -y autoremove
             ;;
         opensuse-*)
@@ -337,7 +349,10 @@ distro_query_package_info() {
         ubuntu-*|debian-*)
             apt-cache policy "$1"
             ;;
-        fedora-*)
+        amazon-*|centos-7-*)
+            yum info "$1"
+            ;;
+        fedora-*|centos-*)
             dnf info "$1"
             ;;
         opensuse-*)
@@ -346,6 +361,10 @@ distro_query_package_info() {
         arch-*)
             pacman -Si "$1"
             ;;
+        *)
+            echo "ERROR: Unsupported distribution '$SPREAD_SYSTEM'"
+            exit 1
+            ;;
     esac
 }
 
@@ -358,11 +377,22 @@ distro_install_build_snapd(){
         apt install -y --only-upgrade snapd
         mv sources.list.back /etc/apt/sources.list
         apt update
+
         # On trusty we may pull in a new hwe-kernel that is needed to run the
         # snapd tests. We need to reboot to actually run this kernel.
-        if [[ "$SPREAD_SYSTEM" = ubuntu-14.04-* ]] && [ "$SPREAD_REBOOT" = 0 ]; then
+        if os.query is-trusty && [ "$SPREAD_REBOOT" = 0 ]; then
             REBOOT
         fi
+    elif [ -n "$PPA_VALIDATION_NAME" ]; then
+        apt install -y snapd
+        add-apt-repository -y "$PPA_VALIDATION_NAME"
+        apt update
+        apt install -y --only-upgrade snapd
+        add-apt-repository --remove "$PPA_VALIDATION_NAME"
+        apt update
+
+        # Double check that it really comes from the PPA
+        apt show snapd | grep "APT-Sources: http.*ppa.launchpad.net"
     else
         packages=
         case "$SPREAD_SYSTEM" in
@@ -370,9 +400,9 @@ distro_install_build_snapd(){
                 # shellcheck disable=SC2125
                 packages="${GOHOME}"/snapd_*.deb
                 ;;
-            fedora-*)
+            fedora-*|amazon-*|centos-*)
                 # shellcheck disable=SC2125
-                packages="${GOHOME}"/snap-confine*.rpm\ "${GOPATH}"/snapd*.rpm
+                packages="${GOHOME}"/snap-confine*.rpm\ "${GOPATH%%:*}"/snapd*.rpm
                 ;;
             opensuse-*)
                 # shellcheck disable=SC2125
@@ -380,7 +410,7 @@ distro_install_build_snapd(){
                 ;;
             arch-*)
                 # shellcheck disable=SC2125
-                packages="${GOHOME}"/snapd*.pkg.tar.xz
+                packages="${GOHOME}"/snapd*.pkg.tar.*
                 ;;
             *)
                 exit 1
@@ -390,10 +420,41 @@ distro_install_build_snapd(){
         # shellcheck disable=SC2086
         distro_install_local_package $packages
 
-        if [[ "$SPREAD_SYSTEM" == arch-* ]]; then
+        case "$SPREAD_SYSTEM" in
+            fedora-*|centos-*)
+                # We need to wait until the man db cache is updated before do daemon-reexec
+                # Otherwise the service fails and the system will be degraded during tests executions
+                for i in $(seq 20); do
+                    if ! systemctl is-active run-*.service; then
+                        break
+                    fi
+                    sleep .5
+                done
+
+                # systemd caches SELinux policy data and subsequently attempts
+                # to create sockets with incorrect context, this installation of
+                # socket activated snaps fails, see:
+                # https://bugzilla.redhat.com/show_bug.cgi?id=1660141
+                # https://github.com/systemd/systemd/issues/9997
+                systemctl daemon-reexec
+                ;;
+        esac
+
+        if os.query is-arch-linux; then
             # Arch policy does not allow calling daemon-reloads in package
             # install scripts
             systemctl daemon-reload
+
+            # AppArmor policy needs to be reloaded
+            if systemctl show -p ActiveState apparmor.service | MATCH 'ActiveState=active'; then
+                systemctl restart apparmor.service
+            fi
+        fi
+
+        if [[ "$SPREAD_SYSTEM" == opensuse-tumbleweed-* ]]; then
+            # Package installation applies vendor presets only, which leaves
+            # snapd.apparmor disabled.
+            systemctl enable --now snapd.apparmor.service
         fi
 
         # On some distributions the snapd.socket is not yet automatically
@@ -412,7 +473,7 @@ distro_get_package_extension() {
         ubuntu-*|debian-*)
             echo "deb"
             ;;
-        fedora-*|opensuse-*)
+        fedora-*|opensuse-*|amazon-*|centos-*)
             echo "rpm"
             ;;
         arch-*)
@@ -428,6 +489,7 @@ pkg_dependencies_ubuntu_generic(){
         automake
         autotools-dev
         build-essential
+        clang
         curl
         devscripts
         expect
@@ -446,50 +508,93 @@ pkg_dependencies_ubuntu_generic(){
         pkg-config
         python3-docutils
         udev
+        udisks2
+        upower
         uuid-runtime
         "
 }
 
 pkg_dependencies_ubuntu_classic(){
     echo "
+        avahi-daemon
         cups
         dbus-x11
+        fontconfig
         gnome-keyring
         jq
         man
+        nfs-kernel-server
         printer-driver-cups-pdf
+        python3-dbus
+        python3-gi
         python3-yaml
         upower
         weston
         xdg-user-dirs
         xdg-utils
+        zsh
         "
 
     case "$SPREAD_SYSTEM" in
         ubuntu-14.04-*)
-            echo "
-                linux-image-extra-$(uname -r)
-                "
+                pkg_linux_image_extra
             ;;
         ubuntu-16.04-32)
             echo "
-                linux-image-extra-$(uname -r)
+                dbus-user-session
+                gccgo-6
+                evolution-data-server
+                fwupd
+                gnome-online-accounts
+                packagekit
                 "
+                pkg_linux_image_extra
             ;;
         ubuntu-16.04-64)
             echo "
+                dbus-user-session
+                evolution-data-server
+                fwupd
                 gccgo-6
+                gnome-online-accounts
                 kpartx
                 libvirt-bin
-                linux-image-extra-$(uname -r)
+                packagekit
                 qemu
                 x11-utils
                 xvfb
                 "
+                pkg_linux_image_extra
             ;;
-        ubuntu-17.10-64)
+        ubuntu-18.04-64)
             echo "
-                linux-image-extra-4.13.0-16-generic
+                dbus-user-session
+                gccgo-8
+                evolution-data-server
+                fwupd
+                packagekit
+                qemu-utils
+                "
+            ;;
+        ubuntu-20.04-64)
+            echo "
+                evolution-data-server
+                gccgo-9
+                packagekit
+                qemu-utils
+                shellcheck
+                "
+            ;;
+        ubuntu-20.10-64)
+            echo "
+                qemu-utils
+                "
+            ;;
+        ubuntu-21.04-64)
+            echo "
+                dbus-user-session
+                golang
+                qemu-utils
                 "
             ;;
         ubuntu-*)
@@ -499,76 +604,168 @@ pkg_dependencies_ubuntu_classic(){
             ;;
         debian-*)
             echo "
+                autopkgtest
+                debootstrap
+                dbus-user-session
+                eatmydata
+                evolution-data-server
+                fwupd
+                gcc-multilib
+                libc6-dev-i386
+                linux-libc-dev
                 net-tools
+                packagekit
+                sbuild
+                schroot
                 "
             ;;
     esac
 }
 
+pkg_linux_image_extra (){
+    if apt-cache show "linux-image-extra-$(uname -r)" > /dev/null 2>&1; then
+        echo "linux-image-extra-$(uname -r)";
+    else
+        if apt-cache show "linux-modules-extra-$(uname -r)" > /dev/null 2>&1; then
+            echo "linux-modules-extra-$(uname -r)";
+        else
+            echo "cannot find a matching kernel modules package";
+            exit 1;
+        fi;
+    fi
+}
+
 pkg_dependencies_ubuntu_core(){
     echo "
-        linux-image-extra-$(uname -r)
         pollinate
         "
+        pkg_linux_image_extra
 }
 
 pkg_dependencies_fedora(){
     echo "
+        clang
         curl
         dbus-x11
+        evolution-data-server
         expect
+        fontconfig
+        fwupd
         git
         golang
         jq
         iptables-services
         man
-        mock
         net-tools
+        nmap-ncat
+        nfs-utils
+        PackageKit
+        python3-yaml
+        python3-dbus
+        python3-gobject
         redhat-lsb-core
         rpm-build
+        udisks2
+        upower
         xdg-user-dirs
+        xdg-utils
+        strace
+        zsh
+        "
+}
+
+pkg_dependencies_amazon(){
+    echo "
+        curl
+        dbus-x11
+        expect
+        fontconfig
+        fwupd
+        git
+        golang
+        grub2-tools
+        jq
+        iptables-services
+        man
+        nc
+        net-tools
+        nfs-utils
+        PackageKit
+        system-lsb-core
+        rpm-build
+        xdg-user-dirs
+        xdg-utils
+        udisks2
+        upower
+        zsh
         "
 }
 
 pkg_dependencies_opensuse(){
     echo "
         apparmor-profiles
+        clang
         curl
+        evolution-data-server
         expect
+        fontconfig
+        fwupd
         git
         golang-packaging
         jq
         lsb-release
         man
+        nfs-kernel-server
+        PackageKit
+        python3-yaml
         netcat-openbsd
         osc
+        udisks2
+        upower
         uuidd
-        xdg-utils
         xdg-user-dirs
+        xdg-utils
+        zsh
         "
 }
 
 pkg_dependencies_arch(){
     echo "
-    curl
     base-devel
+    bash-completion
+    clang
+    curl
+    evolution-data-server
+    expect
+    fontconfig
+    fwupd
+    git
     go
     go-tools
+    jq
     libseccomp
     libcap
+    libx11
+    man
+    net-tools
+    nfs-utils
+    openbsd-netcat
+    packagekit
+    python
     python-docutils
-    xfsprogs
+    python-dbus
+    python-gobject
+    python3-yaml
     squashfs-tools
     shellcheck
-    python
-    jq
-    git
-    openbsd-netcat
+    strace
+    udisks2
+    upower
     xdg-user-dirs
-    expect
-    libx11
-    bash-completion
-    net-tools
+    xdg-utils
+    xfsprogs
+    apparmor
+    zsh
     "
 }
 
@@ -582,7 +779,10 @@ pkg_dependencies(){
             pkg_dependencies_ubuntu_generic
             pkg_dependencies_ubuntu_classic
             ;;
-        fedora-*)
+        amazon-*|centos-7-*)
+            pkg_dependencies_amazon
+            ;;
+        fedora-*|centos-*)
             pkg_dependencies_fedora
             ;;
         opensuse-*)
@@ -599,7 +799,7 @@ pkg_dependencies(){
 install_pkg_dependencies(){
     pkgs=$(pkg_dependencies)
     # shellcheck disable=SC2086
-    distro_install_package $pkgs
+    distro_install_package --no-install-recommends $pkgs
 }
 
 # upgrade distribution and indicate if reboot is needed by outputting 'reboot'
